package com.bergerkiller.bukkit.nolagg;

import java.util.HashMap;
import java.util.logging.Level;

import org.bukkit.craftbukkit.CraftWorld;

import com.bergerkiller.bukkit.common.Operation;
import com.bergerkiller.bukkit.common.utils.CommonUtil;
import com.bergerkiller.bukkit.common.utils.MathUtil;
import com.bergerkiller.bukkit.common.utils.WorldUtil;

import net.minecraft.server.Block;
import net.minecraft.server.Chunk;
import net.minecraft.server.ChunkCoordIntPair;
import net.minecraft.server.EntityPlayer;
import net.minecraft.server.WorldServer;

public class ChunkFixLightingOperation implements ThreadOperation {

	private static HashMap<org.bukkit.Chunk, ChunkFixLightingOperation> running = new HashMap<org.bukkit.Chunk, ChunkFixLightingOperation>();
	
	public static int processed = 0;
	
	private final int x, z;
	private final WorldServer world;
	private final org.bukkit.Chunk key;
	private final Chunk chunk;
	public ChunkFixLightingOperation(org.bukkit.Chunk chunk) {
		this(WorldUtil.getNative(chunk));
	}
	public ChunkFixLightingOperation(final Chunk chunk) {
		this.world = (WorldServer) chunk.world;
		this.x = chunk.x;
		this.z = chunk.z;
		this.chunk = chunk;
		this.key = chunk.bukkitChunk;
		//cancel
		if (this.key != null) {
			ChunkFixLightingOperation op = running.put(this.key, this);
			if (op != null) op.isCancelled = true;
		} else {
			this.isCancelled = true;
		}
	}
	
	private boolean isCancelled = false;
	
	@Override
	public void start() {

	}
	
	private Chunk getChunk(final int x, final int z) {
		return this.world.chunkProviderServer.chunks.get(x, z);
	}
	private int getLightLevel(int x, final int y, int z) {
		if (y <= 0 || y >= 127) return 0;
		if (x >= 0 && z >= 0 && x < 16 && z < 16) {
			return this.chunk.h.a(x, y, z);
		}
		
	    Chunk chunk = this.getChunk(this.x + (x >> 4), this.z + (z >> 4));
		if (chunk == null) return 0;
		x -= (chunk.x - this.x) << 4;
		z -= (chunk.z - this.z) << 4;
		return chunk.h.a(x, y, z);
	}
	private boolean fixBlockLighting(final int x, final int y, final int z) {
		final int typeid = this.chunk.getTypeId(x, y, z);
		if (!Block.o[typeid]) {
			final int factor = Math.max(1, Block.lightBlock[typeid]);
			final int light = this.getLightLevel(x, y, z);
			//actual editing here
			int newlight = light + factor;
			newlight = Math.max(newlight, getLightLevel(x - 1, y, z));
			newlight = Math.max(newlight, getLightLevel(x + 1, y, z));
			newlight = Math.max(newlight, getLightLevel(x, y, z - 1));
			newlight = Math.max(newlight, getLightLevel(x, y, z + 1));
			newlight = Math.max(newlight, getLightLevel(x, y - 1, z));
			newlight = Math.max(newlight, getLightLevel(x, y + 1, z));
			newlight -= factor;
			//when exceeding in loops, just pick the highest value
			if (newlight > light) {
				this.chunk.h.a(x, y, z, newlight);
				return true;
			}
		}
		return false;
	}
	
	public String toString() {
		CraftWorld world = this.world.getWorld();
		if (world == null) {
			return "[" + this.x + "/" + this.z + "]";
		} else {
			return "[" + this.x + "/" + this.z + "/" + world.getName() + "]";
		}
	}

	@Override
	public void run() {
		if (this.isCancelled) return;
		boolean haserror = true;
		int loops = -1;
		int x, y, z;
		while (haserror) {
			if (loops > 20) {
				NoLagg.plugin.log(Level.WARNING, "Failed to fix all lighting issues in chunk " + this.toString());
				break;
			}
			haserror = false;
			loops++;
			for (x = 0; x < 16; x++) {
				for (z = 0; z < 16; z++) {
					for (y = this.chunk.b(x, z); y > 0; --y) {
						if (fixBlockLighting(x, y, z)) {
							haserror = true;
						}
					}
				}
			}
		}
	}

	@Override
	public void end() {
		if (this.isCancelled) return;
		//transfer block data
		Chunk chunk = this.getChunk(this.x, this.z);
		if (chunk != null) {
			final ChunkCoordIntPair pair = new ChunkCoordIntPair(chunk.x, chunk.z);
			new Operation() {
				public void run() {
					this.doPlayers(world);
				}
				@SuppressWarnings("unchecked")
				public void handle(EntityPlayer ep) {
					if (Math.abs(pair.x - MathUtil.locToChunk(ep.locX)) > CommonUtil.view) return;
					if (Math.abs(pair.z - MathUtil.locToChunk(ep.locZ)) > CommonUtil.view) return;
					ep.chunkCoordIntPairQueue.add(0, pair);
				}
			};			
		}
		running.remove(this.key);
		processed++;
	}

}
